package polyglot.ext.carcov;

import polyglot.types.SemanticException;
import java.lang.reflect.*;
import java.util.*;

/**
 * This class composes two factories, and for any method call on the factory,
 * calls the method on both factories, and decides which newly manufactured
 * object to return.
 *
 * Restrictions:
 *  - A FactoryComposer object can be cast to all and only the
 *    interfaces implemented by either factory.
 */
public class FactoryComposer implements InvocationHandler {

    private Object factory1;
    private Object factory2;

    /**
     * Create a new proxy, which implements all the interfaces implemented by
     * factory1 and factory2. i.e. if factory1 or factory2 is an instance
     * of some interface, then the object returned by this method will also
     * be an instance of that interface.
     *
     * The proxy class is defined in the class loader of the class of factory1.
     *
     */
    public static Object newInstance(Object factory1, Object factory2) {
        return newInstance(factory1.getClass().getClassLoader(), factory1, factory2);
    }

    /**
     * Create a new proxy, which implements all the interfaces implemented by
     * factory1 and factory2. i.e. if factory1 or factory2 is an instance
     * of some interface, then the object returned by this method will also
     * be an instance of that interface.
     *
     * The proxy class is defined in the class loader specified.
     *
     */
    public static Object newInstance(ClassLoader cl, Object factory1, Object factory2) {
        return Proxy.newProxyInstance(
            cl,
            getAllInterfaces(factory1, factory2),
            new FactoryComposer(factory1, factory2));
    }

    /**
     * This method navigates it way up the class heirarchy of both
     * objects, and returns an array of all of the interfaces these
     * objects implement.
     */
    private static Class[] getAllInterfaces(Object obj1, Object obj2) {
        Set s = new HashSet();
        Class c = obj1.getClass();
        while (c != null) {
            s.addAll(Arrays.asList(c.getInterfaces()));
            c = c.getSuperclass();
        }

        c = obj2.getClass();
        while (c != null) {
            s.addAll(Arrays.asList(c.getInterfaces()));
            c = c.getSuperclass();
        }

        return (Class[])s.toArray(new Class[] {});
    }

    /**
     * Private constructor, instances can only be created by
     * the static method newInstance
     */
    private FactoryComposer(Object fact1, Object fact2) {
        this.factory1 = fact1;
        this.factory2 = fact2;
    }

    /**
     * Handle the invocations of methods on the proxy.
     *
     * Delegate the invocation to one or both of the factories, and return
     * one of the objects created.
     *
     * No guarantees are made as to which factory or factories will have the
     * method invoked.
     *
     * If one of the factories throws an exception during the invocation of
     * its method, then the exception will be rethrown by this method.
     *
     * @throws SemanticException if the two factories return objects with
     *        incomparable types.
     */
    public Object invoke(Object proxy, Method m, Object[] args)
        throws Throwable
    {
        // A method on one of the interfaces has been invoked. Delegate the
        // call to the two factories, and then decide which object to return.
        try {
            Object o1, o2;
            try {
                o1 = m.invoke(factory1, args);
            }
            catch (IllegalArgumentException e) {
                // the method doesn't exist on the factory factory1
                // Let's take our chances with factory2...
                return m.invoke(factory2, args);
            }

            try {
                o2 = m.invoke(factory2, args);
            }
            catch (IllegalArgumentException e) {
                // the method doesn't exist on the factory factory1
                // Let's take our chances with factory2...
                return o1;
            }


            // at this point, both factories returned something.
            // If one of them was null, then return the other.
            // Otherwise, if one of them is a subclass of the other,
            // return the subclass object.
            // Otherwise, the two objects are incompatible, and the factories
            // cannot be composed.
            if (o1 == null) {
                return o2;
            } else if (o2 == null) {
                return o1;
            } else if (o1.getClass().isInstance(o2) || o1.getClass().equals(o2.getClass())) {
                // o2 is a subclass of the class of o1, so it is more specialised.
                // return it in preference to o1.
                return o2;
            }else if (o2.getClass().isInstance(o1)) {
                // o1 is a subclass of the class of o2, so it is more specialised.
                // return it in preference to o2.
                return o1;
            } else {
                // o1 and o2 are both non-null, and
                throw new SemanticException("Incompatible fatories: " +
                            factory1.getClass().getName() + " and " +
                            factory2.getClass().getName() +
                            " return objects of type " +
                            o1.getClass().getName() + " and " +
                            o2.getClass().getName() +
                            " respectively for the method " +
                            m.getName());
            }
        } catch (InvocationTargetException e) {
            throw e.getTargetException();
        } catch (Exception e) {
            throw new RuntimeException("unexpected invocation exception: " +
                                       e.getMessage());
        }
    }
}
